#include "main.h"


void refresh_leds(void);

volatile unsigned int execution_speed_1;
volatile unsigned int execution_speed_2;



signed int decode_int(volatile char * * str, volatile const char * end, signed int max)
{
	if (max < 0)
		return -1;
	if (end <= *str)
		return -1;
	signed int retval = 0;
	while (*str < end)
	{
		char c = **str;
		(*str)++;
		if (c == ':')
			return retval;
		unsigned char d = c - '0';
		if (d > 9)
			return -1;
		if ((retval > 3276) || ((retval == 3276) && (c >= 8)))
			return -1;
		retval *= 10;
		retval += d;
		if (retval > max)
			return -1;
	}
	return retval;
}

//unsigned int pwm[4][8];

int main(void)
{
	DDRA  = 0b00000000;
	PORTA = 0b11111111;
	DDRB  = 0b00000000;
	PORTB = 0b11111111;
	DDRC  = 0b00000000;
	PORTC = 0b11111111;
	DDRD  = 0b00000000;
	PORTD = 0b11111111;
	sei();
	initialize_leds();
	initialize_uart();
	initialize_timer();
	
	bool was_sleeping = 0;
	send_response(11);
	volatile unsigned int execution_speed_2;
	while (1)
	{
		unsigned int start = TCNT1;
		volatile struct recv_buffer * buffer = get_next_data();
		if (buffer == NULL)
		{
			refresh_leds();
		}
		else
		{
			if (was_sleeping)
			{
				was_sleeping = 0;
				send_response(8);
			}
			else if (buffer->size == 0)
			{
				send_response(1);
			}
			else if (buffer->size == -1)
			{
				send_response(5);
			}
			else if (buffer->size >= 1)
			{
				volatile char * str = buffer->data;
				volatile char * end = buffer->data + buffer->size;
				char c = *str;
				str++;
				switch (c)
				{
					case 'A':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int addr = decode_int(&str, end, 3);
						unsigned int value = decode_int(&str, end, 15);
						if ((addr > 3) || (value > 15) || (str != end))
						{
							send_response(2);
							break;
						}
						set_dac(value, addr);
						send_response(0);
						break;
					}
					case 'L':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int addr = decode_int(&str, end, 7);
						if (addr >= 8)
						{
							send_response(2);
							break;
						}
						unsigned int luminosity[4];
						for (unsigned char q = 0; q < 4; q++)
						{
							luminosity[q] = decode_int(&str, end, 1023);
							if (luminosity[q] > 1023)
							{
								send_response(2);
								goto break_switch;
							}
						}
						if (str != end)
						{
							send_response(2);
							break;
						}
						ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
						{
							for (unsigned char q = 0; q < 4; q++)
								leds[addr].luminosity[q] = luminosity[q];
						}
						send_response(0);
						break;
					}
					case 'B':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int addr = decode_int(&str, end, 7);
						if (addr >= 8)
						{
							send_response(2);
							break;
						}
						unsigned int duty = decode_int(&str, end, 255);
						if (duty > 255)
						{
							send_response(2);
							break;
						}
						ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
						{
							leds[addr].duty = (unsigned char)duty;
							leds[addr].gamma = 0;
						}
						send_response(0);
						break;
					}
					case 'I':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int addr = decode_int(&str, end, 7);
						if (addr >= 8)
						{
							send_response(2);
							break;
						}
						unsigned int gamma = decode_int(&str, end, 255);
						if (gamma > 255)
						{
							send_response(2);
							break;
						}
						ATOMIC_BLOCK(ATOMIC_RESTORESTATE) //maybe optional - is it already essentially atomic?
						{
							leds[addr].gamma = gamma;
						}
						send_response(0);
						break;
					}
					case 'P':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int addr = decode_int(&str, end, 7);
						if (addr >= 8)
						{
							send_response(2);
							break;
						}
						unsigned int phase = decode_int(&str, end, 255);
						if (phase > 255)
						{
							send_response(2);
							break;
						}
						ATOMIC_BLOCK(ATOMIC_RESTORESTATE) //maybe optional - is it already essentially atomic?
						{
							leds[addr].phase_offset = phase;
						}
						send_response(0);
						break;
					}
					case 'T':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int phase = decode_int(&str, end, 255);
						if (phase > 255)
						{
							send_response(2);
							break;
						}
						ATOMIC_BLOCK(ATOMIC_RESTORESTATE) //maybe optional - is it already essentially atomic?
						{
							theta = phase;
						}
						send_response(0);
						break;
					}
					case 'F':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int period = decode_int(&str, end, 32767);
						if ((period > 32767) || (period < 10))
						{
							send_response(2);
							break;
						}
						set_timer_period(period);
						send_response(0);
						break;
					}
					case 'N':
					{
						send_response(0);
						break;
					}
					case 'S':
					{
						ATOMIC_BLOCK(ATOMIC_FORCEON)
						{
							set_multi_dac(0, 0, 0, 0);
								for (unsigned char i = 0; i < 8; i++)
							for (unsigned int q = 0; q < 8; q++)
							{
								leds[q].luminosity[0] = 0;
								leds[q].luminosity[1] = 0;
								leds[q].luminosity[2] = 0;
								leds[q].luminosity[3] = 0;
								leds[q].gamma = 0;
								leds[q].duty = 0;
								leds[q].phase_offset = 0;
								set_multi_pwm(q, 0, 0, 0, 0);
							}
							cancel_timer();
						}
						send_response(9);
						while ((UCSRA & 0b00100000) == 0)
							;
						power_down();
						initialize_timer();
						was_sleeping = 1;
						break;
					}
					case 'X':
					{
						send_int(10);
						send_char(':');
						send_int(execution_speed_1);
						send_char(':');
						send_int(execution_speed_2);
						send_char('\n');
						break;
					}
					case 'E':
					{
						if ((str >= end) || (*str != ':'))
						{
							send_response(2);
							break;
						}
						str++;
						unsigned int retval = decode_int(&str, end, 255);
						send_response(retval);
						break;
					}
					default:
					{
						send_response(1);
					}
				}
			}
			else
			{
				send_response(6);
			}
			break_switch:
			buffer->valid = 0;
			execution_speed_2 = TCNT1 - start;
		}
	}
	send_response(99);
	return main();
}




//gamma could be 10 bit, if it was faster
/*
for GAMMA in 1.8 2.0 2.2 ; do
 echo -n '0'
 for i in `seq 1 255` ; do
  echo -n ', '
  echo "x=(e(${GAMMA}*l(${i}/255))*255+0.5);scale=0;x/1" | bc -l | tr -d '\n'
 done
 echo ''
done
*/
static const unsigned char gamma1_8[256] PROGMEM = {
0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 16, 16, 17, 17, 18, 18, 19, 19, 20, 21, 21, 22, 22, 23, 24, 24, 25, 26, 26, 27, 28, 28, 29, 30, 30, 31, 32, 32, 33, 34, 35, 35, 36, 37, 38, 38, 39, 40, 41, 41, 42, 43, 44, 45, 46, 46, 47, 48, 49, 50, 51, 52, 53, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 86, 87, 88, 89, 90, 91, 92, 93, 95, 96, 97, 98, 99, 100, 102, 103, 104, 105, 107, 108, 109, 110, 111, 113, 114, 115, 116, 118, 119, 120, 122, 123, 124, 126, 127, 128, 129, 131, 132, 134, 135, 136, 138, 139, 140, 142, 143, 145, 146, 147, 149, 150, 152, 153, 154, 156, 157, 159, 160, 162, 163, 165, 166, 168, 169, 171, 172, 174, 175, 177, 178, 180, 181, 183, 184, 186, 188, 189, 191, 192, 194, 195, 197, 199, 200, 202, 204, 205, 207, 208, 210, 212, 213, 215, 217, 218, 220, 222, 224, 225, 227, 229, 230, 232, 234, 236, 237, 239, 241, 243, 244, 246, 248, 250, 251, 253, 255
};
static const unsigned char gamma2_0[256] PROGMEM = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 23, 23, 24, 24, 25, 26, 26, 27, 28, 28, 29, 30, 30, 31, 32, 32, 33, 34, 35, 35, 36, 37, 38, 38, 39, 40, 41, 42, 42, 43, 44, 45, 46, 47, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 84, 85, 86, 87, 88, 89, 91, 92, 93, 94, 95, 97, 98, 99, 100, 102, 103, 104, 105, 107, 108, 109, 111, 112, 113, 115, 116, 117, 119, 120, 121, 123, 124, 126, 127, 128, 130, 131, 133, 134, 136, 137, 139, 140, 142, 143, 145, 146, 148, 149, 151, 152, 154, 155, 157, 158, 160, 162, 163, 165, 166, 168, 170, 171, 173, 175, 176, 178, 180, 181, 183, 185, 186, 188, 190, 192, 193, 195, 197, 199, 200, 202, 204, 206, 207, 209, 211, 213, 215, 217, 218, 220, 222, 224, 226, 228, 230, 232, 233, 235, 237, 239, 241, 243, 245, 247, 249, 251, 253, 255
};
static const unsigned char gamma2_2[256] PROGMEM = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 22, 22, 23, 23, 24, 25, 25, 26, 26, 27, 28, 28, 29, 30, 30, 31, 32, 33, 33, 34, 35, 35, 36, 37, 38, 39, 39, 40, 41, 42, 43, 43, 44, 45, 46, 47, 48, 49, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 73, 74, 75, 76, 77, 78, 79, 81, 82, 83, 84, 85, 87, 88, 89, 90, 91, 93, 94, 95, 97, 98, 99, 100, 102, 103, 105, 106, 107, 109, 110, 111, 113, 114, 116, 117, 119, 120, 121, 123, 124, 126, 127, 129, 130, 132, 133, 135, 137, 138, 140, 141, 143, 145, 146, 148, 149, 151, 153, 154, 156, 158, 159, 161, 163, 165, 166, 168, 170, 172, 173, 175, 177, 179, 181, 182, 184, 186, 188, 190, 192, 194, 196, 197, 199, 201, 203, 205, 207, 209, 211, 213, 215, 217, 219, 221, 223, 225, 227, 229, 231, 234, 236, 238, 240, 242, 244, 246, 248, 251, 253, 255
};


void refresh_leds(void)
{
	unsigned int start = TCNT1;
	unsigned int theta_copy = theta; //use same value of theta for each LED - don't allow changing right now
	
	for (unsigned char i = 0; i < 8; i++)
	{
		unsigned char phase = (theta_copy + leds[i].phase_offset);
		if (leds[i].gamma == 0)
		{
			if (phase <= leds[i].duty)
				set_multi_pwm(i, leds[i].luminosity[0], leds[i].luminosity[1], leds[i].luminosity[2], leds[i].luminosity[3]);
			else
				set_multi_pwm(i, 0, 0, 0, 0);
		}
		else
		{
			if (phase > 128)
				phase = 256 - phase;
			unsigned char r;
			unsigned char g;
			unsigned char b;
			if (leds[i].gamma == 1)
			{
				r = (unsigned char)(((leds[i].luminosity[0] >> 1) * phase) >> 8);
				g = (unsigned char)(((leds[i].luminosity[0] >> 1) * phase) >> 8);
				b = (unsigned char)(((leds[i].luminosity[0] >> 1) * phase) >> 8);
			}
			else
			{
				const unsigned char * cur_gamma;
				switch (leds[i].gamma)
				{
					case 2:
						cur_gamma = gamma1_8;
						break;
					case 3:
						cur_gamma = gamma2_0;
						break;
					case 4:
					default:
						cur_gamma = gamma2_2;
						break;
				}
				r = pgm_read_byte(cur_gamma + (unsigned char)(((leds[i].luminosity[0] >> 1) * phase) >> 8));
				g = pgm_read_byte(cur_gamma + (unsigned char)(((leds[i].luminosity[1] >> 1) * phase) >> 8));
				b = pgm_read_byte(cur_gamma + (unsigned char)(((leds[i].luminosity[2] >> 1) * phase) >> 8));
			}
			set_multi_pwm(i, r << 2, g << 2, b << 2, 0);
		}
	}
	
	execution_speed_1 = TCNT1 - start;
}

